#!/usr/bin/perl

use strict;
use warnings;
no warnings 'uninitialized', 'once';
use LWP::UserAgent;
use IO::File;
use FindBin;

$|++;

our ($VERSION) = "1.02";
our ($PROGRAM) = 'PCCC Raptor FAST for Linux';
our ($COPYRIGHT) = 'Copyright (C) 2014 Peregrine Computer Consultants Corporation. All rights reserved.';
our ($VERBOSE) = 0;

my (%FORM, $FORM);
$FORM = \%FORM;

print "$PROGRAM v$VERSION\n$COPYRIGHT\n\n";

if (-e "$FindBin::Bin/config") {
  print "Loading Configuration... \n";
  require "$FindBin::Bin/config";
  if ($main::token eq '') {
    print "Error.  You must have a valid Raptor FAST Token defined in your config file!\nExiting.\n\n";
    exit;
  }
} else {
  print "Error Loading Configuration! You must copy config.dist.linux to config and edit it for your installation.\n";
  exit;
}

#SANITIZE VARIABLES
$main::virtusertable ||= '/etc/mail/virtusertable';
$main::localhostnames ||= '/etc/mail/local-host-names';
$main::aliases ||= "/etc/mail/aliases";
$main::passwd ||= "/etc/passwd";

&main($FORM, token=>$main::token, verbose=>$main::VERBOSE);

exit;

sub main {
  my ($FORM, %params) = @_;
  my ($fh, $filename, %catchall_domains, %catchall_error_domains, $domain, @emails, $email, $emails, %domains, @manual_domains, %catchall_redir_domains, $redir_domain);
  
  $fh = new IO::File;
  $filename = $main::virtusertable;

  if ($fh->open("< $filename")) {
    while (<$fh>) {
      chomp($_);
      #ADD THE DOMAIN TO THE HASH OF ALL DOMAINS - FOR COMMENTS, ERRORS AND SPACE
      if ($_ !~ /(^#|\serror:|^\s*$)/i) {
        $_ =~ m/\@(.*?)\s/;
        $domain = $1;
        $domains{$domain}++; 
        print "DEBUG: Hash of All Domains in VirtUserTable $domain\n" if $VERBOSE > 0;
      }
 
      #ERROR LINE FOR DOMAIN
      if ($_ =~ m/^@.*\s*error:/) {
        print "DEBUG: Catch-All for Domain with Error - Activated [due to $_]\n" if $VERBOSE > 0;

        #THE DOMAIN IS GOOD TO USE - IT HAS A CATCH-ALL FOR ERROR
        $_ =~ m/^\@(.*?)\s.*$/;
        $domain = $1;
        print "DEBUG: Catch-all with Error [$domain]\n" if $VERBOSE > 0;
        $catchall_error_domains{$domain}++;
      #CATCH ALL WITH LOCAL DELIVERY
      } elsif ($_ =~ m/^@.*\s*%1\s*$/) {
        print "DEBUG: Catch-All for Domain with Local Delivery - Ignoring.  Will be handled below with Aliases and Local Account Tests [due to $_]\n" if $VERBOSE > 0;
      #CATCH ALL WITH REDIRECTION - NEED TO CHECK
      } elsif ($_ =~ m/^@.*\s*%.*@/) {
        print "DEBUG: Catch-All for Domain with Redirection - Undetermined [due to $_]\n" if $VERBOSE > 0;
        #THE DOMAIN REDIRECTS EMAIL - NEED TO CHECK IT AFTER THE PROCESSING IS DONE FOR VIRTUSERTABLE
        $_ =~ m/^\@(.*?)\s(.*)$/;
        $domain = $1;
        $redir_domain = $2;
        $redir_domain =~ s/^.*@(.*)$/$1/;
        print "DEBUG: Catch-all with Redirection [$domain redirs to $redir_domain]\n" if $VERBOSE > 0;
        $catchall_redir_domains{$domain} = $redir_domain;
      #CATCH ALL FOR DOMAIN
      } elsif ($_ =~ m/^\@/) {
        print "DEBUG: Catch-All for Domain without Error - Activated [due to $_]\n" if $VERBOSE > 0;

        #THE DOMAIN ALLOWS ALL EMAILS - IT HAS A CATCH-ALL MAKING ALL EMAILS VALID
        $_ =~ m/^\@(.*?)\s.*$/;
        $domain = $1;
        print "DEBUG: Catch-all without Error [$domain]\n" if $VERBOSE > 0;
        $catchall_domains{$domain}++;
      #NOT A COMMENT, BLANK OR AN ERROR MESSAGE
      } elsif ($_ !~ /(^#|\serror:|^\s*$)/i) {
        $_ =~ s/\s+.*$//;
        $email = $_;
        print "DEBUG: $email\n" if $VERBOSE > 0;
        push (@emails, $email); 
      }
    }
    close ($fh);

    #CHECK DOMAINS WITH REDIRECTION IF THERE IS DOMAIN THEY REDIR TO IN ANOTHER LIST
    foreach $domain (keys %catchall_redir_domains) {
      print "DEBUG: Check if $domain redirecting to $catchall_redir_domains{$domain} is in Catch-All lists.\n" if $VERBOSE > 0;

      print "DEBUG: ".($catchall_domains{$catchall_redir_domains{$domain}} + $catchall_error_domains{$catchall_redir_domains{$domain}})."\n" if $VERBOSE > 0;
      if ($catchall_domains{$catchall_redir_domains{$domain}} > 0) {
        $catchall_domains{$domain}++;
      }
      if ($catchall_error_domains{$catchall_redir_domains{$domain}} > 0) {
        $catchall_error_domains{$domain}++;
      }

      undef $catchall_redir_domains{$domain};
    }
  }
  print "\n";

  #ADD CATCHALL AS @domainname  AND CREATE HASH
  print "Catch-All without Error Domains:\n";
  foreach $domain (keys %catchall_domains) {
    print "\t$domain\n";
    $emails .= "\@$domain\n";
  }
  print "\n";

  #CREATE CATCH-ALL WITH ERROR HASH
  print "Catch-All with Error Domains:\n";
  foreach $domain (keys %catchall_error_domains) {
    print "\t$domain\n";
  }
  print "\n";

  #READ /etc/mail/local-host-names
  $filename = $main::localhostnames;

  if ($fh->open("< $filename")) {
    while (<$fh>) {
      #IF NOT A COMMENT OR BLANK
      if ($_ !~ /(^#|^\s*$)/i) {
        $domain = $_;
        #STRIP ALL TRAILING WHITESPACE
        $domain =~ s/\s*$//g;
        chomp($domain);
        $domains{$domain}++;
        print "DEBUG: $domain\n" if $VERBOSE > 0;
      }
    }
    close($fh);
  }

  #DO WE HAVE ANY DOMAINS IN %DOMAINS THAT DON'T EXIST IN A CATCHALL DOMAIN LIST?  IF SO, WE HAVE MORE CHECKS TO DO:
  foreach $domain (keys %domains) {
    unless ($catchall_domains{$domain} + $catchall_error_domains{$domain}) {
      print "DEBUG: This domain doesn't exist in the catch-all lists $domain\n" if $VERBOSE > 0;
      print "DEBUG: $catchall_domains{$domain}\n$catchall_error_domains{$domain}\n" if $VERBOSE > 0;
    }
    print "DEBUG: ".($catchall_domains{$domain} + $catchall_error_domains{$domain})."\n" if $VERBOSE > 0;

    if (($catchall_domains{$domain} + $catchall_error_domains{$domain}) < 1) {
      #DOMAIN DOESN'T HAVE A CATCHALL.  NEED TO PROCESS ACCESS AND ACCOUNTS
      push @manual_domains, $domain;
    }
  }

  if (scalar(@manual_domains) > 0) {
    my %addresses;

    #READ /etc/mail/aliases
    $filename = $main::aliases;

    if ($fh->open("< $filename")) {
      while (<$fh>) {
        #IF NOT A COMMENT OR BLANK
        if ($_ !~ /(^#|^\s*$)/i) {
          $_ = m/^(.+?):/;
          $email = $1;
          $addresses{$email}++;
          print "DEBUG: $email\n" if $VERBOSE > 0;
        }
      }
      close($fh);
    }

    #READ /etc/passwd
    $filename = $main::passwd;

    if ($fh->open("< $filename")) {
      while (<$fh>) {
        #IF NOT A COMMENT OR BLANK
        if ($_ !~ /(^#|^\s*$)/i) {
          $_ = m/^(.+?)\:/;
          $email = $1;
          $addresses{$email}++;
          print "DEBUG: $email\n" if $VERBOSE > 0;
        }
      }
      close($fh);
    }

    #ADD ADDRESSES + MANUAL DOMAINS TO EMAIL LIST
    print "Manual Domains (Relying on Aliases and Accounts or External Redirection):\n";
    foreach $domain (@manual_domains) {
      print "\t$domain\n";
      foreach $email (keys %addresses) {
        $emails .= "$email\@$domain\n";
      }
    }
    print "\n";
  }

  #CHECK EACH EMAIL 
  foreach $email (@emails) {
    $email =~ m/\@(.*)$/;
    $domain = $1;

    print "DEBUG: Domain for Email $email is $domain\n" if $VERBOSE > 0;
    print "DEBUG: Catchall_domains ".$catchall_domains{uc($domain)}."\n" if $VERBOSE > 0;

    unless ($catchall_domains{uc($domain)}) {
      $emails .= "$email\n";
    }
  }

  print "DEBUG:\n$emails\n" if $VERBOSE > 0;
  #REPORT THE RESULTS TO THE RAPTOR SERVER
  &submit_emails_to_raptor($FORM, emails=>$emails, token=>$params{'token'});
}

sub submit_emails_to_raptor {
  my ($FORM, %params) = @_;
  my ($user_agent, $request, $result);

  #REPORT THE RESULTS TO THE RAPTOR SERVER
  $user_agent = new LWP::UserAgent;
  $request = new HTTP::Request("POST", "https://www.pccc.com/raptor_FAST_submit.cgim");

  $request->content_type("application/x-www-form-urlencoded");

  #URL ENCODE THE INFORMATION
  $params{'emails'} =~ s/([^a-zA-Z0-9_.-])/uc sprintf("%%%02x",ord($1))/eg;

  $request->content("token=$params{'token'}&emails=$params{'emails'}");
  $result = $user_agent->request($request);
  if ($result->is_error) { 
    print "Error submitting LDAP information to Raptor.\n";
    print "  Message: ".$result->status_line."\n";
  } else {
    print "Raptor submission succesful.\n";
  }
}
